@@ -8,26 +8,26 @@
 #include <sys/ioctl.h>
 #include <errno.h>
 
-#include "sg_include.h"
 #include "path_state.h"
 #include "checkers.h"
 
+#include "../libmultipath/sg_include.h"
+
 #define INQUIRY_CMD     0x12
 #define INQUIRY_CMDLEN  6
 #define HEAVY_CHECK_COUNT       10
 
 struct emc_clariion_checker_context {
-	int fd;
 	int run_count;
 	char wwn[16];
 	unsigned wwn_set;
 };
 
-int emc_clariion(char *devt, char *msg, void **context)
+int emc_clariion(int fd, char *msg, void **context)
 {
-	unsigned char sense_buffer[256];
-	unsigned char sb[128];
-	unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 0, 0xC0, 0,
+	unsigned char sense_buffer[256] = { 0, };
+	unsigned char sb[128] = { 0, };
+	unsigned char inqCmdBlk[INQUIRY_CMDLEN] = {INQUIRY_CMD, 1, 0xC0, 0,
 						sizeof(sb), 0};
 	struct sg_io_hdr io_hdr;
 	struct emc_clariion_checker_context * ctxt = NULL;
@@ -61,13 +61,10 @@ int emc_clariion(char *devt, char *msg, void **context)
 		/* do stuff */
 	}
 
-	if (!ctxt->fd) {
-		if (devnode(CREATE_NODE, devt)) {
-			ret = -1;
-			goto out;
-		}
-		ctxt->fd = devnode(OPEN_NODE, devt);
-		devnode(UNLINK_NODE, devt);
+	if (fd <= 0) {
+		MSG("no usable fd");
+		ret = -1;
+		goto out;
 	}
 	memset(&io_hdr, 0, sizeof (struct sg_io_hdr));
 	io_hdr.interface_id = 'S';
@@ -80,7 +77,7 @@ int emc_clariion(char *devt, char *msg, void **context)
 	io_hdr.sbp = sb;
 	io_hdr.timeout = 60000;
 	io_hdr.pack_id = 0;
-	if (ioctl(ctxt->fd, SG_IO, &io_hdr) < 0) {
+	if (ioctl(fd, SG_IO, &io_hdr) < 0) {
 		MSG("emc_clariion_checker: sending query command failed");
 		ret = PATH_DOWN;
 		goto out;
@@ -131,7 +128,7 @@ int emc_clariion(char *devt, char *msg, void **context)
 	 * pointing somewhere else.
 	 */
 	if (context && ctxt->wwn_set) {
-		if (memcmp(ctxt->wwn, &sense_buffer[10], 16)) {
+		if (memcmp(ctxt->wwn, &sense_buffer[10], 16) != 0) {
 			MSG("emc_clariion_checker: Logical Unit WWN has changed!");
 			ret = PATH_DOWN;
 			goto out;
@@ -149,9 +146,8 @@ out:
 	 * caller told us he doesn't want to keep the context :
 	 * free it
 	 */
-	if (!context) {
-		close(ctxt->fd);
+	if (!context)
 		free(ctxt);
-	}
+
 	return(ret);
 }
