@@ -10,14 +10,19 @@ get_checker_id (char * str)
 		return TUR;
 	if (0 == strncmp(str, "readsector0", 11))
 		return READSECTOR0;
-
+ 	if (0 == strncmp(str, "directio", 8))
+		return DIRECTIO;
+ 	if (0 == strncmp(str, "emc_clariion", 12))
+		return EMC_CLARIION;
+ 	if (0 == strncmp(str, "hp_sw", 5))
+		return HP_SW;
 	return -1;
 }
 
 extern void * 
 get_checker_addr (int id)
 {
-	int (*checker) (char *, char *, void **);
+	int (*checker) (int, char *, void **);
 
 	switch (id) {
 	case TUR:
@@ -26,9 +31,15 @@ get_checker_addr (int id)
 	case READSECTOR0:
 		checker = &readsector0;
 		break;
+	case DIRECTIO:
+		checker = &directio;
+		break;
 	case EMC_CLARIION:
 		checker = &emc_clariion;
 		break;
+	case HP_SW:
+		checker = &hp_sw;
+		break;
 	default:
 		checker = NULL;
 		break;
@@ -48,9 +59,15 @@ get_checker_name (char * str, int id)
 	case READSECTOR0:
 		s = "readsector0";
 		break;
+	case DIRECTIO:
+		s = "directio";
+		break;
 	case EMC_CLARIION:
 		s = "emc_clariion";
 		break;
+	case HP_SW:
+		s = "hp_sw";
+		break;
 	default:
 		s = "undefined";
 		break;
