@@ -21,7 +21,7 @@ multipath_tool_handler(vector strvec)
 	if (!conf->multipath)
 		return 1;
 
-	return push_callout(conf->multipath);
+	return 0;
 }
 
 static int
@@ -31,6 +31,7 @@ polling_interval_handler(vector strvec)
 
 	buff = VECTOR_SLOT(strvec, 1);
 	conf->checkint = atoi(buff);
+	conf->max_checkint = conf->checkint << 2;
 
 	return 0;
 }
@@ -81,7 +82,7 @@ def_getuid_callout_handler(vector strvec)
 	if (!conf->default_getuid)
 		return 1;
 	
-	return push_callout(conf->default_getuid);
+	return 0;
 }
 
 static int
@@ -92,7 +93,12 @@ def_prio_callout_handler(vector strvec)
 	if (!conf->default_getprio)
 		return 1;
 	
-	return push_callout(conf->default_getprio);
+	if (!strncmp(conf->default_getprio, "none", 4)) {
+		FREE(conf->default_getprio);
+		conf->default_getprio = NULL;
+	}
+		
+	return 0;
 }
 
 static int
@@ -122,6 +128,44 @@ def_minio_handler(vector strvec)
 	return 0;
 }
 
+static int
+def_weight_handler(vector strvec)
+{
+	char * buff;
+
+	buff = set_value(strvec);
+
+	if (!buff)
+		return 1;
+
+	if (strlen(buff) == 10 &&
+	    !strcmp(buff, "priorities"))
+		conf->rr_weight = RR_WEIGHT_PRIO;
+
+	FREE(buff);
+
+	return 0;
+}
+
+static int
+default_failback_handler(vector strvec)
+{
+	char * buff;
+
+	buff = set_value(strvec);
+
+	if (!strncmp(buff, "manual", 6))
+		conf->pgfailback = -FAILBACK_MANUAL;
+	else if (!strncmp(buff, "immediate", 9))
+		conf->pgfailback = -FAILBACK_IMMEDIATE;
+	else
+		conf->pgfailback = atoi(buff);
+
+	FREE(buff);
+
+	return 0;
+}
+
 /*
  * blacklist block handlers
  */
@@ -245,7 +289,7 @@ hw_getuid_callout_handler(vector strvec)
 	if (!hwe->getuid)
 		return 1;
 
-	return push_callout(hwe->getuid);
+	return 0;
 }
 
 static int
@@ -329,7 +373,58 @@ prio_callout_handler(vector strvec)
 	if (!hwe->getprio)
 		return 1;
 
-	return push_callout(hwe->getprio);
+	if (!strncmp(hwe->getprio, "none", 4)) {
+		FREE(hwe->getprio);
+		hwe->getprio = NULL;
+	}
+
+	return 0;
+}
+
+static int
+hw_failback_handler(vector strvec)
+{
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+	char * buff;
+
+	if (!hwe)
+		return 1;
+
+	buff = set_value(strvec);
+
+	if (!strncmp(buff, "manual", 6))
+		hwe->pgfailback = -FAILBACK_MANUAL;
+	else if (!strncmp(buff, "immediate", 9))
+		hwe->pgfailback = -FAILBACK_IMMEDIATE;
+	else
+		hwe->pgfailback = atoi(buff);
+
+	FREE(buff);
+
+	return 0;
+}
+
+static int
+hw_weight_handler(vector strvec)
+{
+	struct hwentry * hwe = VECTOR_LAST_SLOT(conf->hwtable);
+	char * buff;
+
+	if (!hwe)
+		return 1;
+
+	buff = set_value(strvec);
+
+	if (!buff)
+		return 1;
+
+	if (strlen(buff) == 10 &&
+	    !strcmp(buff, "priorities"))
+		hwe->rr_weight = RR_WEIGHT_PRIO;
+
+	FREE(buff);
+
+	return 0;
 }
 
 /*
@@ -433,6 +528,52 @@ mp_selector_handler(vector strvec)
 	return 0;
 }
 
+static int
+mp_failback_handler(vector strvec)
+{
+	struct mpentry * mpe = VECTOR_LAST_SLOT(conf->mptable);
+	char * buff;
+
+	if (!mpe)
+		return 1;
+
+	buff = set_value(strvec);
+
+	if (!strncmp(buff, "manual", 6))
+		mpe->pgfailback = -FAILBACK_MANUAL;
+	else if (!strncmp(buff, "immediate", 9))
+		mpe->pgfailback = -FAILBACK_IMMEDIATE;
+	else
+		mpe->pgfailback = atoi(buff);
+
+	FREE(buff);
+
+	return 0;
+}
+
+static int
+mp_weight_handler(vector strvec)
+{
+	struct mpentry * mpe = VECTOR_LAST_SLOT(conf->mptable);
+	char * buff;
+
+	if (!mpe)
+		return 1;
+
+	buff = set_value(strvec);
+
+	if (!buff)
+		return 1;
+
+	if (strlen(buff) == 10 &&
+	    !strcmp(buff, "priorities"))
+		mpe->rr_weight = RR_WEIGHT_PRIO;
+
+	FREE(buff);
+
+	return 0;
+}
+
 vector
 init_keywords(void)
 {
@@ -447,7 +588,9 @@ init_keywords(void)
 	install_keyword("default_getuid_callout", &def_getuid_callout_handler);
 	install_keyword("default_prio_callout", &def_prio_callout_handler);
 	install_keyword("default_features", &def_features_handler);
+	install_keyword("failback", &default_failback_handler);
 	install_keyword("rr_min_io", &def_minio_handler);
+	install_keyword("rr_weight", &def_weight_handler);
 	
 	install_keyword_root("devnode_blacklist", &blacklist_handler);
 	install_keyword("devnode", &ble_handler);
@@ -465,6 +608,8 @@ init_keywords(void)
 	install_keyword("features", &hw_features_handler);
 	install_keyword("hardware_handler", &hw_handler_handler);
 	install_keyword("prio_callout", &prio_callout_handler);
+	install_keyword("failback", &hw_failback_handler);
+	install_keyword("rr_weight", &hw_weight_handler);
 	install_sublevel_end();
 
 	install_keyword_root("multipaths", &multipaths_handler);
@@ -474,6 +619,8 @@ init_keywords(void)
 	install_keyword("alias", &alias_handler);
 	install_keyword("path_grouping_policy", &mp_pgpolicy_handler);
 	install_keyword("path_selector", &mp_selector_handler);
+	install_keyword("failback", &mp_failback_handler);
+	install_keyword("rr_weight", &mp_weight_handler);
 	install_sublevel_end();
 
 	return keywords;
