@@ -2,6 +2,8 @@
 #include <unistd.h>
 #include <fcntl.h>
 #include <sys/ioctl.h>
+#include <sys/stat.h>
+#include <errno.h>
 
 #include <sysfs/dlist.h>
 #include <sysfs/libsysfs.h>
@@ -21,17 +23,43 @@
 #define readattr(a,b) \
 	sysfs_read_attribute_value(a, b, sizeof(b))
 
+struct path *
+store_pathinfo (vector pathvec, vector hwtable, char * devname, int flag)
+{
+	struct path * pp;
+
+	pp = alloc_path();
+
+	if (!pp)
+		return NULL;
+
+	if(safe_sprintf(pp->dev, "%s", devname)) {
+		condlog(0, "pp->dev too small");
+		goto out;
+	}
+	if (pathinfo(pp, hwtable, flag))
+		goto out;
+
+	if (store_path(pathvec, pp))
+		goto out;
+
+	return pp;
+out:
+	free_path(pp);
+	return NULL;
+}
+
 int
 path_discovery (vector pathvec, struct config * conf, int flag)
 {
 	struct sysfs_directory * sdir;
 	struct sysfs_directory * devp;
 	char path[FILE_NAME_SIZE];
-	struct path * curpath;
+	struct path * pp;
 	int r = 1;
 
 	if(safe_sprintf(path, "%s/block", sysfs_path)) {
-		fprintf(stderr, "path too small\n");
+		condlog(0, "path too small");
 		exit(1);
 	}
 	sdir = sysfs_open_directory(path);
@@ -43,36 +71,30 @@ path_discovery (vector pathvec, struct config * conf, int flag)
 
 		if(safe_sprintf(path, "%s/block/%s/device", sysfs_path,
 				devp->name)) {
-			fprintf(stderr, "path too small\n");
+			condlog(0, "path too small");
 			exit(1);
 		}
 				
 		if (!filepresent(path))
 			continue;
 
-		curpath = find_path_by_dev(pathvec, devp->name);
+		pp = find_path_by_dev(pathvec, devp->name);
 
-		if (!curpath) {
-			curpath = alloc_path();
-
-			if (!curpath)
+		if (!pp) {
+			/*
+			 * new path : alloc, store and fetch info
+			 * the caller wants
+			 */
+			if (!store_pathinfo(pathvec, conf->hwtable,
+					   devp->name, flag))
 				goto out;
-
-			if (store_path(pathvec, curpath)) {
-				free_path(curpath);
-				goto out;
-			}
-			if(safe_sprintf(curpath->dev, "%s", devp->name)) {
-				fprintf(stderr, "curpath->dev too small\n");
-				exit(1);
-			}
-			devinfo(curpath, conf->hwtable, DI_ALL);
 		} else {
 			/*
-			 * path already known,
+			 * path already known :
 			 * refresh only what the caller wants
 			 */
-			devinfo(curpath, conf->hwtable, flag);
+			if (pathinfo(pp, conf->hwtable, flag))
+				goto out;
 		}
 	}
 	r = 0;
@@ -81,6 +103,26 @@ out:
 	return r;
 }
 
+#define WAIT_MAX_SECONDS 5
+#define WAIT_LOOP_PER_SECOND 5
+
+static int
+wait_sysfs_attr (char * filename)
+{
+	int loop;
+	struct stat stats;
+	
+	loop = WAIT_MAX_SECONDS * WAIT_LOOP_PER_SECOND;
+	
+	while (--loop) {
+		if (stat(filename, &stats) == 0)
+			return 0;
+
+		usleep(1000 * 1000 / WAIT_LOOP_PER_SECOND);
+	}
+	return 1;
+}	
+			
 #define declare_sysfs_get_str(fname, fmt) \
 extern int \
 sysfs_get_##fname (char * sysfs_path, char * dev, char * buff, int len) \
@@ -89,8 +131,12 @@ sysfs_get_##fname (char * sysfs_path, char * dev, char * buff, int len) \
 	char attr_buff[SYSFS_PATH_SIZE]; \
 	int attr_len; \
 \
-	if(safe_sprintf(attr_path, fmt, sysfs_path, dev)) \
+	if (safe_sprintf(attr_path, fmt, sysfs_path, dev)) \
 		return 1; \
+\
+	if (wait_sysfs_attr(attr_path)) \
+		return 1; \
+\
 	if (0 > sysfs_read_attribute_value(attr_path, attr_buff, sizeof(attr_buff))) \
 		return 1; \
 \
@@ -109,50 +155,77 @@ declare_sysfs_get_str(rev, "%s/block/%s/device/rev");
 declare_sysfs_get_str(dev, "%s/block/%s/dev");
 
 #define declare_sysfs_get_val(fname, fmt) \
-extern unsigned long  \
+extern unsigned long long  \
 sysfs_get_##fname (char * sysfs_path, char * dev) \
 { \
 	char attr_path[SYSFS_PATH_SIZE]; \
 	char attr_buff[SYSFS_PATH_SIZE]; \
+	int r; \
+	unsigned long long val; \
 \
-	if(safe_sprintf(attr_path, fmt, sysfs_path, dev)) \
+	if (safe_sprintf(attr_path, fmt, sysfs_path, dev)) \
 		return 0; \
+\
+	if (wait_sysfs_attr(attr_path)) \
+		return 0; \
+\
 	if (0 > sysfs_read_attribute_value(attr_path, attr_buff, sizeof(attr_buff))) \
 		return 0; \
 \
-	return strtoul(attr_buff, NULL, 0); \
+	r = sscanf(attr_buff, "%llu\n", &val); \
+\
+	if (r != 1) \
+		return 0; \
+	else \
+		return (val); \
 }
 
 declare_sysfs_get_val(size, "%s/block/%s/size");
 
+/*
+ * udev might be slow creating node files : wait
+ */
 static int
 opennode (char * dev, int mode)
 {
 	char devpath[FILE_NAME_SIZE];
 	int fd;
+	int loop;
 
 	if (safe_sprintf(devpath, "%s/%s", conf->udev_dir, dev)) {
-		fprintf(stderr, "devpath too small\n");
+		condlog(0, "devpath too small");
 		return -1;
 	}
-	fd = open(devpath, mode);
 
-	if (fd <= 0)
-		condlog(0, "open(%s) failed", devpath);
+	loop = WAIT_MAX_SECONDS * WAIT_LOOP_PER_SECOND;
 	
-	return fd;
+	while (--loop) {
+		fd = open(devpath, mode);
+
+		if (fd <= 0 && errno != ENOENT) {
+			condlog(3, "open error (%s)\n", strerror(errno));
+			return fd;
+		}
+		if (fd > 0)
+			return fd;
+
+		usleep(1000 * 1000 / WAIT_LOOP_PER_SECOND);
+	}
+	condlog(0, "failed to open %s", devpath);
+	return -1;
 }
 
-#if 0
 int
-get_claimed(int fd)
+get_claimed(char * devname)
 {
-	/*
-	 * FIXME : O_EXCL always fails ?
-	 */
+	int fd = opennode(devname, O_EXCL);
+
+	if (fd < 0)
+		return 1;
+
+	close(fd);
 	return 0;
 }	
-#endif
 
 extern int
 devt2devname (char *devname, char *devt)
@@ -165,7 +238,7 @@ devt2devname (char *devname, char *devt)
 	int len;
 
 	if(safe_sprintf(block_path, "%s/block", sysfs_path)) {
-		fprintf(stderr, "block_path too small\n");
+		condlog(0, "block_path too small");
 		exit(1);
 	}
 	sdir = sysfs_open_directory(block_path);
@@ -174,7 +247,7 @@ devt2devname (char *devname, char *devt)
 	dlist_for_each_data (sdir->subdirs, devp, struct sysfs_directory) {
 		if(safe_sprintf(attr_path, "%s/%s/dev",
 				block_path, devp->name)) {
-			fprintf(stderr, "attr_path too small\n");
+			condlog(0, "attr_path too small");
 			exit(1);
 		}
 		sysfs_read_attribute_value(attr_path, attr_value,
@@ -189,7 +262,7 @@ devt2devname (char *devname, char *devt)
 		    strncmp(attr_value, devt, len) == 0) {
 			if(safe_sprintf(attr_path, "%s/%s",
 					block_path, devp->name)) {
-				fprintf(stderr, "attr_path too small\n");
+				condlog(0, "attr_path too small");
 				exit(1);
 			}
 			sysfs_get_name_from_path(attr_path, devname,
@@ -267,15 +340,48 @@ get_serial (char * str, int fd)
 		len = buff[3];
 		if (len > 0) {
 			memcpy(str, buff + 4, len);
-			buff[len] = '\0';
+			str[len] = '\0';
 		}
 		return 1;
 	}
         return 0;
 }
 
-extern int
-sysfs_devinfo(struct path * curpath)
+static void
+sysfs_get_bus (char * sysfs_path, struct path * curpath)
+{
+	struct sysfs_device *sdev;
+	char attr_path[FILE_NAME_SIZE];
+	char attr_buff[FILE_NAME_SIZE];
+
+	curpath->bus = SYSFS_BUS_UNDEF;
+
+	/*
+	 * This is ugly : we should be able to do a simple
+	 * get_link("%s/block/%s/device/bus", ...) but it just
+	 * won't work
+	 */
+	if(safe_sprintf(attr_path, "%s/block/%s/device",
+			sysfs_path, curpath->dev)) {
+		condlog(0, "attr_path too small");
+		return;
+	}
+
+	if (0 > sysfs_get_link(attr_path, attr_buff, sizeof(attr_buff)))
+		return;
+
+	sdev = sysfs_open_device_path(attr_buff);
+
+	if (!strncmp(sdev->bus, "scsi", 4))
+		curpath->bus = SYSFS_BUS_SCSI;
+	else if (!strncmp(sdev->bus, "ide", 3))
+		curpath->bus = SYSFS_BUS_IDE;
+
+	return;
+}
+
+static int
+scsi_sysfs_pathinfo (struct path * curpath)
 {
 	char attr_path[FILE_NAME_SIZE];
 	char attr_buff[FILE_NAME_SIZE];
@@ -283,40 +389,34 @@ sysfs_devinfo(struct path * curpath)
 	if (sysfs_get_vendor(sysfs_path, curpath->dev,
 			     curpath->vendor_id, SCSI_VENDOR_SIZE))
 		return 1;
+
 	condlog(3, "vendor = %s", curpath->vendor_id);
 
 	if (sysfs_get_model(sysfs_path, curpath->dev,
 			    curpath->product_id, SCSI_PRODUCT_SIZE))
 		return 1;
+
 	condlog(3, "product = %s", curpath->product_id);
 
 	if (sysfs_get_rev(sysfs_path, curpath->dev,
 			  curpath->rev, SCSI_REV_SIZE))
 		return 1;
+
 	condlog(3, "rev = %s", curpath->rev);
 
-	if (sysfs_get_dev(sysfs_path, curpath->dev,
-			  curpath->dev_t, BLK_DEV_SIZE))
-		return 1;
-	condlog(3, "dev_t = %s", curpath->dev_t);
-
-	curpath->size = sysfs_get_size(sysfs_path, curpath->dev);
-
-	if (curpath->size == 0)
-		return 1;
-	condlog(3, "size = %lu", curpath->size);
-
 	/*
 	 * host / bus / target / lun
 	 */
 	if(safe_sprintf(attr_path, "%s/block/%s/device",
 			sysfs_path, curpath->dev)) {
-		fprintf(stderr, "attr_path too small\n");
+		condlog(0, "attr_path too small");
 		return 1;
 	}
 	if (0 > sysfs_get_link(attr_path, attr_buff, sizeof(attr_buff)))
 		return 1;
+	
 	basename(attr_buff, attr_path);
+
 	sscanf(attr_path, "%i:%i:%i:%i",
 			&curpath->sg_id.host_no,
 			&curpath->sg_id.channel,
@@ -337,7 +437,7 @@ sysfs_devinfo(struct path * curpath)
 			curpath->sg_id.host_no,
 			curpath->sg_id.channel,
 			curpath->sg_id.scsi_id)) {
-		fprintf(stderr, "attr_path too small\n");
+		condlog(0, "attr_path too small");
 		return 1;
 	}
 	if (0 <= readattr(attr_path, attr_buff) && strlen(attr_buff) > 0)
@@ -348,6 +448,44 @@ sysfs_devinfo(struct path * curpath)
 	return 0;
 }
 
+static int
+common_sysfs_pathinfo (struct path * curpath)
+{
+
+	sysfs_get_bus(sysfs_path, curpath);
+	condlog(3, "bus = %i", curpath->bus);
+
+	if (sysfs_get_dev(sysfs_path, curpath->dev,
+			  curpath->dev_t, BLK_DEV_SIZE))
+		return 1;
+
+	condlog(3, "dev_t = %s", curpath->dev_t);
+
+	curpath->size = sysfs_get_size(sysfs_path, curpath->dev);
+
+	if (curpath->size == 0)
+		return 1;
+
+	condlog(3, "size = %llu", curpath->size);
+
+	return 0;
+}
+
+extern int
+sysfs_pathinfo(struct path * curpath)
+{
+	if (common_sysfs_pathinfo(curpath))
+		return 1;
+
+	if (curpath->bus == SYSFS_BUS_UNDEF)
+		return 0;
+	else if (curpath->bus == SYSFS_BUS_SCSI)
+		if (scsi_sysfs_pathinfo(curpath))
+			return 1;
+
+	return 0;
+}
+
 static int
 apply_format (char * string, char * cmd, struct path * pp)
 {
@@ -427,8 +565,19 @@ apply_format (char * string, char * cmd, struct path * pp)
 	return 0;
 }
 
+static int
+scsi_ioctl_pathinfo (struct path * pp, int mask)
+{
+	if (mask & DI_SERIAL) {
+		get_serial(pp->serial, pp->fd);
+		condlog(3, "serial = %s", pp->serial);
+	}
+
+	return 0;
+}
+
 extern int
-devinfo (struct path *pp, vector hwtable, int mask)
+pathinfo (struct path *pp, vector hwtable, int mask)
 {
 	char buff[CALLOUT_MAX_SIZE];
 	char prio[16];
@@ -438,28 +587,25 @@ devinfo (struct path *pp, vector hwtable, int mask)
 	/*
 	 * fetch info available in sysfs
 	 */
-	if (mask & DI_SYSFS && sysfs_devinfo(pp))
+	if (mask & DI_SYSFS && sysfs_pathinfo(pp))
 		return 1;
 
 	/*
 	 * then those not available through sysfs
 	 */
+	if (mask & DI_CLAIMED) {
+		pp->claimed = get_claimed(pp->dev);
+		condlog(3, "claimed = %i", pp->claimed);
+	}
 	if (pp->fd <= 0)
 		pp->fd = opennode(pp->dev, O_RDONLY);
 
 	if (pp->fd <= 0)
 		return 1;
 
-	if (mask & DI_SERIAL) {
-		get_serial(pp->serial, pp->fd);
-		condlog(3, "serial = %s", pp->serial);
-	}
-#if 0
-	if (mask & DI_CLAIMED) {
-		pp->claimed = get_claimed(pp->fd);
-		condlog(3, "claimed = %i", pp->claimed);
-	}
-#endif
+	if (pp->bus == SYSFS_BUS_SCSI)
+		if (scsi_ioctl_pathinfo(pp, mask))
+			return 1;
 
 	/* get and store hwe pointer */
 	pp->hwe = find_hwe(hwtable, pp->vendor_id, pp->product_id);
@@ -467,9 +613,10 @@ devinfo (struct path *pp, vector hwtable, int mask)
 	/*
 	 * get path state, no message collection, no context
 	 */
-	select_checkfn(pp);
-
 	if (mask & DI_CHECKER) {
+		if (!pp->checkfn)
+			select_checkfn(pp);
+
 		pp->state = pp->checkfn(pp->fd, NULL, NULL);
 		condlog(3, "state = %i", pp->state);
 	}
@@ -478,9 +625,12 @@ devinfo (struct path *pp, vector hwtable, int mask)
 	 * get path prio
 	 */
 	if (mask & DI_PRIO) {
-		select_getprio(pp);
+		if (!pp->getprio)
+			select_getprio(pp);
 
-		if (apply_format(pp->getprio, &buff[0], pp)) {
+		if (!pp->getprio) {
+			pp->priority = 1;
+		} else if (apply_format(pp->getprio, &buff[0], pp)) {
 			condlog(0, "error formatting prio callout command");
 			pp->priority = -1;
 		} else if (execute_program(buff, prio, 16)) {
@@ -496,7 +646,8 @@ devinfo (struct path *pp, vector hwtable, int mask)
 	 * get path uid
 	 */
 	if (mask & DI_WWID && !strlen(pp->wwid)) {
-		select_getuid(pp);
+		if (!pp->getuid)
+			select_getuid(pp);
 
 		if (apply_format(pp->getuid, &buff[0], pp)) {
 			condlog(0, "error formatting uid callout command");
