@@ -11,68 +11,10 @@
 #include "memory.h"
 #include "structs.h"
 #include "util.h"
+#include "debug.h"
 
 #define WORD_SIZE 64
 
-#if DEBUG
-#define dbg(format, arg...) fprintf(stderr, format "\n", ##arg)
-#else
-#define dbg(format, arg...) do {} while(0)
-#endif
-
-extern struct path *
-find_path (vector pathvec, char * dev_t)
-{
-	int i;
-	struct path * pp;
-
-	vector_foreach_slot (pathvec, pp, i)
-		if (!strcmp_chomp(pp->dev_t, dev_t))
-			return pp;
-
-	dbg("path %s not found in pathvec\n", dev_t);
-	return NULL;
-}
-
-static int
-get_word (char * sentence, char ** word)
-{
-	char * p;
-	int len;
-	int skip = 0;
-	
-	while (*sentence ==  ' ') {
-		sentence++;
-		skip++;
-	}
-	if (*sentence == '\0')
-		return 0;
-
-	p = sentence;
-
-	while (*p !=  ' ' && *p != '\0')
-		p++;
-
-	len = (int) (p - sentence);
-
-	if (!word)
-		return skip + len;
-
-	*word = zalloc(len + 1);
-
-	if (!*word) {
-		fprintf(stderr, "get_word : oom\n");
-		return 0;
-	}
-	strncpy(*word, sentence, len);
-//	dbg("*word = %s, len = %i", *word, len);
-
-	if (*p == '\0')
-		return 0;
-
-	return skip + len;
-}
-
 static int
 merge_words (char ** dst, char * word, int space)
 {
@@ -80,7 +22,7 @@ merge_words (char ** dst, char * word, int space)
 	int len;
 
 	len = strlen(*dst) + strlen(word) + space;
-	*dst = realloc(*dst, len + 1);
+	*dst = REALLOC(*dst, len + 1);
 
 	if (!*dst)
 		return 1;
@@ -100,18 +42,18 @@ merge_words (char ** dst, char * word, int space)
 	return 0;
 }
 
-extern void
+extern int
 disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 {
 	char * word;
 	char * p;
 	int i, j, k;
-	int num_features;
-	int num_hwhandler;
-	int num_pg;
-	int num_pg_args;
-	int num_paths;
-	int num_paths_args;
+	int num_features = 0;
+	int num_hwhandler = 0;
+	int num_pg = 0;
+	int num_pg_args = 0;
+	int num_paths = 0;
+	int num_paths_args = 0;
 	struct path * pp;
 	struct pathgroup * pgp;
 
@@ -121,42 +63,74 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 	 * features
 	 */
 	p += get_word(p, &mpp->features);
+
+	if (!mpp->features)
+		return 1;
+
 	num_features = atoi(mpp->features);
 
 	for (i = 0; i < num_features; i++) {
 		p += get_word(p, &word);
-		merge_words(&mpp->features, word, 1);
-		free(word);
+
+		if (!word)
+			return 1;
+
+		if (merge_words(&mpp->features, word, 1)) {
+			FREE(word);
+			return 1;
+		}
+		FREE(word);
 	}
 
 	/*
 	 * hwhandler
 	 */
 	p += get_word(p, &mpp->hwhandler);
+
+	if (!mpp->hwhandler)
+		return 1;
+
 	num_hwhandler = atoi(mpp->hwhandler);
 
 	for (i = 0; i < num_hwhandler; i++) {
 		p += get_word(p, &word);
-		merge_words(&mpp->hwhandler, word, 1);
-		free(word);
+
+		if (!word)
+			return 1;
+
+		if (merge_words(&mpp->hwhandler, word, 1)) {
+			FREE(word);
+			return 1;
+		}
+		FREE(word);
 	}
 
 	/*
 	 * nb of path groups
 	 */
 	p += get_word(p, &word);
+
+	if (!word)
+		return 1;
+
 	num_pg = atoi(word);
-	free(word);
+	FREE(word);
 
 	if (num_pg > 0 && !mpp->pg)
 		mpp->pg = vector_alloc();
 	
+	if (!mpp->pg)
+		return 1;
 	/*
 	 * first pg to try
 	 */
 	p += get_word(p, &word);
+
+	if (!word)
+		goto out;
+
 	mpp->nextpg = atoi(word);
-	free(word);
+	FREE(word);
 
 	for (i = 0; i < num_pg; i++) {
 		/*
@@ -165,13 +139,25 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 
 		if (!mpp->selector) {
 			p += get_word(p, &mpp->selector);
+
+			if (!mpp->selector)
+				goto out;
+
 			/*
 			 * selector args
 			 */
 			p += get_word(p, &word);
+
+			if (!word)
+				goto out;
+
 			num_pg_args = atoi(word);
-			merge_words(&mpp->selector, word, 1);
-			free(word);
+			
+			if (merge_words(&mpp->selector, word, 1)) {
+				FREE(word);
+				goto out1;
+			}
+			FREE(word);
 		} else {
 			p += get_word(p, NULL);
 			p += get_word(p, NULL);
@@ -183,46 +169,83 @@ disassemble_map (vector pathvec, char * params, struct multipath * mpp)
 		/*
 		 * paths
 		 */
-		pgp = zalloc(sizeof(struct pathgroup));
-		pgp->paths = vector_alloc();
-		vector_alloc_slot(mpp->pg);
-		vector_set_slot(mpp->pg, pgp);
+		pgp = alloc_pathgroup();
+		
+		if (!pgp)
+			goto out;
+
+		if (store_pathgroup(mpp->pg, pgp))
+			goto out;
 
 		p += get_word(p, &word);
+
+		if (!word)
+			goto out;
+
 		num_paths = atoi(word);
-		free(word);
+		FREE(word);
 
 		p += get_word(p, &word);
+
+		if (!word)
+			goto out;
+
 		num_paths_args = atoi(word);
-		free(word);
+		FREE(word);
 
 		for (j = 0; j < num_paths; j++) {
 			pp = NULL;
 			p += get_word(p, &word);
+
+			if (!word)
+				goto out;
+
 			if (pathvec)
-				pp = find_path(pathvec, word);
+				pp = find_path_by_devt(pathvec, word);
 
 			if (!pp) {
 				pp = alloc_path();
+
+				if (!pp)
+					goto out1;
+
 				strncpy(pp->dev_t, word, BLK_DEV_SIZE);
 			}
-			free(word);
+			FREE(word);
 
-			if (pp) {
-				vector_alloc_slot(pgp->paths);
-				vector_set_slot(pgp->paths, pp);
-				pgp->id ^= (long)pp;
-			}
-			if (pp && !strlen(mpp->wwid))
+			if (store_path(pgp->paths, pp))
+				goto out;
+
+			/*
+			 * Update wwid for multipaths which are not setup
+			 * in the get_dm_mpvec() code path
+			 */
+			if (!strlen(mpp->wwid))
 				strncpy(mpp->wwid, pp->wwid, WWID_SIZE);
 
+			/*
+			 * Update wwid for paths which may not have been
+			 * active at the time the getuid callout was run
+			 */
+			else if (!strlen(pp->wwid))
+				strncpy(pp->wwid, mpp->wwid, WWID_SIZE);
+
+			pgp->id ^= (long)pp;
+			pp->pgindex = i + 1;
+
 			for (k = 0; k < num_paths_args; k++)
 				p += get_word(p, NULL);
 		}
 	}
+	return 0;
+out1:
+	FREE(word);
+out:
+	free_pgvec(mpp->pg, KEEP_PATHS);
+	return 1;
 }
 
-extern void
+extern int
 disassemble_status (char * params, struct multipath * mpp)
 {
 	char * word;
@@ -242,14 +265,22 @@ disassemble_status (char * params, struct multipath * mpp)
 	 * features
 	 */
 	p += get_word(p, &word);
+
+	if (!word)
+		return 1;
+
 	num_feature_args = atoi(word);
-	free(word);
+	FREE(word);
 
 	for (i = 0; i < num_feature_args; i++) {
 		if (i == 1) {
 			p += get_word(p, &word);
+
+			if (!word)
+				return 1;
+
 			mpp->queuedio = atoi(word);
-			free(word);
+			FREE(word);
 			continue;
 		}
 		/* unknown */
@@ -259,8 +290,12 @@ disassemble_status (char * params, struct multipath * mpp)
 	 * hwhandler
 	 */
 	p += get_word(p, &word);
+
+	if (!word)
+		return 1;
+
 	num_hwhandler_args = atoi(word);
-	free(word);
+	FREE(word);
 
 	for (i = 0; i < num_hwhandler_args; i++)
 		p += get_word(p, NULL);
@@ -269,8 +304,12 @@ disassemble_status (char * params, struct multipath * mpp)
 	 * nb of path groups
 	 */
 	p += get_word(p, &word);
+
+	if (!word)
+		return 1;
+
 	num_pg = atoi(word);
-	free(word);
+	FREE(word);
 
 	/*
 	 * next pg to try
@@ -278,7 +317,7 @@ disassemble_status (char * params, struct multipath * mpp)
 	p += get_word(p, NULL);
 
 	if (VECTOR_SIZE(mpp->pg) < num_pg)
-		return;
+		return 1;
 
 	for (i = 0; i < num_pg; i++) {
 		pgp = VECTOR_SLOT(mpp->pg, i);
@@ -286,6 +325,10 @@ disassemble_status (char * params, struct multipath * mpp)
 		 * PG status
 		 */
 		p += get_word(p, &word);
+
+		if (!word)
+			return 1;
+
 		switch (*word) {
 		case 'D':
 			pgp->status = PGSTATE_DISABLED;
@@ -297,10 +340,10 @@ disassemble_status (char * params, struct multipath * mpp)
 			pgp->status = PGSTATE_ENABLED;
 			break;
 		default:
-			pgp->status = PGSTATE_RESERVED;
+			pgp->status = PGSTATE_UNDEF;
 			break;
 		}
-		free(word);
+		FREE(word);
 
 		/*
 		 * undef ?
@@ -308,15 +351,23 @@ disassemble_status (char * params, struct multipath * mpp)
 		p += get_word(p, NULL);
 
 		p += get_word(p, &word);
+
+		if (!word)
+			return 1;
+
 		num_paths = atoi(word);
-		free(word);
+		FREE(word);
 
 		p += get_word(p, &word);
+
+		if (!word)
+			return 1;
+
 		num_pg_args = atoi(word);
-		free(word);
+		FREE(word);
 
 		if (VECTOR_SIZE(pgp->paths) < num_paths)
-			return;
+			return 1;
 
 		for (j = 0; j < num_paths; j++) {
 			pp = VECTOR_SLOT(pgp->paths, j);
@@ -329,6 +380,10 @@ disassemble_status (char * params, struct multipath * mpp)
 			 * path status
 			 */
 			p += get_word(p, &word);
+
+			if (!word)
+				return 1;
+
 			switch (*word) {
 			case 'F':
 				pp->dmstate = PSTATE_FAILED;
@@ -339,13 +394,17 @@ disassemble_status (char * params, struct multipath * mpp)
 			default:
 				break;
 			}
-			free(word);
+			FREE(word);
 			/*
 			 * fail count
 			 */
 			p += get_word(p, &word);
+
+			if (!word)
+				return 1;
+
 			pp->failcount = atoi(word);
-			free(word);
+			FREE(word);
 		}
 		/*
 		 * selector args
@@ -353,4 +412,5 @@ disassemble_status (char * params, struct multipath * mpp)
 		for (j = 0; j < num_pg_args; j++)
 			p += get_word(p, NULL);
 	}
+	return 0;
 }
