@@ -14,14 +14,37 @@
 #define SCSI_PRODUCT_SIZE	17
 #define SCSI_REV_SIZE		5
 
+enum free_path_switch {
+	KEEP_PATHS,
+	FREE_PATHS
+};
+
+enum rr_weight_mode {
+	RR_WEIGHT_UNDEF,
+	RR_WEIGHT_NONE,
+	RR_WEIGHT_PRIO
+};
+
+enum failback_mode {
+	FAILBACK_UNDEF,
+	FAILBACK_MANUAL,
+	FAILBACK_IMMEDIATE
+};
+
+enum sysfs_buses {
+	SYSFS_BUS_UNDEF,
+	SYSFS_BUS_SCSI,
+	SYSFS_BUS_IDE
+};
+
 enum pathstates {
-	PSTATE_RESERVED,
+	PSTATE_UNDEF,
 	PSTATE_FAILED,
 	PSTATE_ACTIVE
 };
 
 enum pgstates {
-	PGSTATE_RESERVED,
+	PGSTATE_UNDEF,
 	PGSTATE_ENABLED,
 	PGSTATE_DISABLED,
 	PGSTATE_ACTIVE
@@ -61,17 +84,22 @@ struct path {
 	char rev[SCSI_REV_SIZE];
 	char serial[SERIAL_SIZE];
 	char tgt_node_name[NODE_NAME_SIZE];
-	unsigned long size;
+	unsigned long long size;
+	unsigned int checkint;
+	unsigned int tick;
 	int state;
+	int bus;
 	int dmstate;
 	int failcount;
-	unsigned int priority;
+	int priority;
 	int claimed;
+	int pgindex;
 	char * getuid;
 	char * getprio;
-	int (*checkfn) (char *, char *, void **);
+	int (*checkfn) (int, char *, void **);
 	void * checker_context;
 	struct multipath * mpp;
+	int fd;
 	
 	/* configlet pointers */
 	struct hwentry * hwe;
@@ -79,23 +107,30 @@ struct path {
 
 struct multipath {
 	char wwid[WWID_SIZE];
-	char * alias;
+	int minor;
 	int pgpolicy;
 	int nextpg;
 	int queuedio;
 	int action;
-	unsigned long size;
+	int pgfailback;
+	int failback_tick;
+	int rr_weight;
+	unsigned long long size;
 	vector paths;
 	vector pg;
 	char params[PARAMS_SIZE];
 	char status[PARAMS_SIZE];
+
+	/* configlet pointers */
+	char * alias;
 	char * selector;
 	char * features;
 	char * hwhandler;
-
-	/* configlet pointers */
 	struct mpentry * mpe;
 	struct hwentry * hwe;
+
+	/* daemon store a data blob for DM event waiter threads */
+	void * waiter;
 };
 
 struct pathgroup {
@@ -106,8 +141,25 @@ struct pathgroup {
 };
 
 struct path * alloc_path (void);
+struct pathgroup * alloc_pathgroup (void);
 struct multipath * alloc_multipath (void);
-void free_multipath (struct multipath *);
+void free_path (struct path *);
+void free_pathvec (vector vec, int free_paths);
+void free_pathgroup (struct pathgroup * pgp, int free_paths);
+void free_pgvec (vector pgvec, int free_paths);
+void free_multipath (struct multipath *, int free_paths);
+void free_multipathvec (vector mpvec, int free_paths);
+
+int store_path (vector pathvec, struct path * pp);
+int store_pathgroup (vector pgvec, struct pathgroup * pgp);
+
 struct multipath * find_mp (vector mp, char * alias);
+struct multipath * find_mp_by_wwid (vector mp, char * wwid);
+struct multipath * find_mp_by_minor (vector mp, int minor);
+	
+struct path * find_path_by_devt (vector pathvec, char * devt);
+struct path * find_path_by_dev (vector pathvec, char * dev);
+
+char sysfs_path[FILE_NAME_SIZE];
 
 #endif
