@@ -1,8 +1,4 @@
 /* 
- * Soft:        Keepalived is a failover program for the LVS project
- *              <www.linuxvirtualserver.org>. It monitor & manipulate
- *              a loadbalanced server pool using multi-layer checks.
- * 
  * Part:        Vector structure manipulation.
  *  
  * Version:     $Id: vector.c,v 1.0.3 2003/05/11 02:28:03 acassen Exp $
@@ -26,45 +22,104 @@
 
 /* 
  * Initialize vector struct.
- * allocalted 'size' slot elements then return vector.
+ * allocated 'size' slot elements then return vector.
  */
 vector
 vector_alloc(void)
 {
-	vector v = (vector) zalloc(sizeof (struct _vector));
+	vector v = (vector) MALLOC(sizeof (struct _vector));
 	return v;
 }
 
 /* allocated one slot */
-void
+void *
 vector_alloc_slot(vector v)
 {
 	v->allocated += VECTOR_DEFAULT_SIZE;
 	if (v->slot)
-		v->slot = realloc(v->slot, sizeof (void *) * v->allocated);
+		v->slot = REALLOC(v->slot, sizeof (void *) * v->allocated);
 	else
-		v->slot = (void *) zalloc(sizeof (void *) * v->allocated);
+		v->slot = (void *) MALLOC(sizeof (void *) * v->allocated);
+
+	if (!v->slot)
+		v->allocated -= VECTOR_DEFAULT_SIZE;
+
+	return v->slot;
 }
 
-void
+void *
 vector_insert_slot(vector v, int slot, void *value)
 {
 	int i;
 	
-	vector_alloc_slot(v);
+	if (!vector_alloc_slot(v))
+		return NULL;
 
-	for (i = v->allocated - 2; i >= slot; i -= VECTOR_DEFAULT_SIZE)
-		v->slot[i+VECTOR_DEFAULT_SIZE] = v->slot[i];
+	for (i = (v->allocated /VECTOR_DEFAULT_SIZE) - 2; i >= slot; i--)
+		v->slot[i + 1] = v->slot[i];
 
-	v->slot[slot*VECTOR_DEFAULT_SIZE] = value;
+	v->slot[slot] = value;
+
+	return v->slot[slot];
 }
-		
+
+int
+find_slot(vector v, void * addr)
+{
+	int i;
+
+	for (i = 0; i < (v->allocated / VECTOR_DEFAULT_SIZE); i++)
+		if (v->slot[i] == addr)
+			return i;
+
+	return -1;
+}
+
+void
+vector_del_slot(vector v, int slot)
+{
+	int i;
+
+	if (!v->allocated)
+		return;
+
+	for (i = slot + 1; i < (v->allocated / VECTOR_DEFAULT_SIZE); i++)
+		v->slot[i-1] = v->slot[i];
+
+	v->allocated -= VECTOR_DEFAULT_SIZE;
+
+	if (!v->allocated) {
+		FREE(v->slot);
+		v->slot = NULL;
+	}
+	else
+		v = REALLOC(v->slot, sizeof (void *) * v->allocated);
+}
+
+void
+vector_repack(vector v)
+{
+	int i;
+
+	if (!v->allocated)
+		return;
+
+	for (i = 0; i < (v->allocated / VECTOR_DEFAULT_SIZE); i++)
+		if (i > 0 && v->slot[i] == NULL)
+			vector_del_slot(v, i--);
+}
+
 /* Free memory vector allocation */
 void
 vector_free(vector v)
 {
-	free(v->slot);
-	free(v);
+	if (!v)
+		return;
+
+	if (v->slot)
+		FREE(v->slot);
+
+	FREE(v);
 }
 
 void
@@ -76,9 +131,9 @@ free_strvec(vector strvec)
 	if (!strvec)
 		return;
 
-	for (i = 0; i < VECTOR_SIZE(strvec); i++)
-		if ((str = VECTOR_SLOT(strvec, i)) != NULL)
-			free(str);
+	vector_foreach_slot (strvec, str, i)
+		if (str)
+			FREE(str);
 
 	vector_free(strvec);
 }
