@@ -19,218 +19,149 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <fcntl.h>
-#include <sys/types.h>
-#include <sys/stat.h>
 #include <string.h>
-#include <signal.h>
-#include <sys/ioctl.h>
 
 #include <parser.h>
 #include <vector.h>
 #include <memory.h>
-#include <sysfs/dlist.h>
-#include <sysfs/libsysfs.h>
 #include <libdevmapper.h>
 #include <devmapper.h>
 #include <checkers.h>
 #include <path_state.h>
-#include <safe_printf.h>
 #include <blacklist.h>
 #include <hwtable.h>
 #include <util.h>
 #include <defaults.h>
 #include <structs.h>
 #include <dmparser.h>
+#include <cache.h>
+#include <config.h>
+#include <propsel.h>
+#include <discovery.h>
+#include <debug.h>
+#include <switchgroup.h>
+#include <sysfs/libsysfs.h>
+#include <print.h>
 
 #include "main.h"
-#include "devinfo.h"
-#include "config.h"
 #include "pgpolicies.h"
 #include "dict.h"
-#include "debug.h"
-#include "propsel.h"
 
-static int
-filepresent (char * run) {
-	struct stat buf;
+/* for column aligned output */
+struct path_layout pl;
 
-	if(!stat(run, &buf))
-		return 1;
-	return 0;
-}
-
-static int
-get_pathvec_sysfs (vector pathvec)
+static char *
+get_refwwid (vector pathvec)
 {
-	struct sysfs_directory * sdir;
-	struct sysfs_directory * devp;
-	struct sysfs_link * linkp;
-	char refwwid[WWID_SIZE];
-	char empty_buff[WWID_SIZE];
+	struct path * pp;
 	char buff[FILE_NAME_SIZE];
-	char path[FILE_NAME_SIZE];
-	char sysfs_path[FILE_NAME_SIZE];
-	struct path * curpath;
+	char * refwwid;
 
-	memset (empty_buff, 0, WWID_SIZE);
-	memset (refwwid, 0, WWID_SIZE);
+	if (conf->dev_type == DEV_NONE)
+		return NULL;
 
-	if (sysfs_get_mnt_path(sysfs_path, FILE_NAME_SIZE)) {
-		fprintf(stderr, "multipath tools need sysfs\n");
-		exit(1);
-	}
-	
-	/*
-	 * if called from /etc/dev.d or pathcheckers, only consider the paths
-	 * that relate to the device pointed by conf->dev
-	 */
 	if (conf->dev_type == DEV_DEVNODE) {
-		dbg("limited scope = %s", conf->dev);
-		curpath = zalloc(sizeof (struct path));
-		basename(conf->dev, curpath->dev);
-
-		if (devinfo(curpath))
-			return 1;
-
-		memcpy(refwwid, curpath->wwid, WWID_SIZE);
-		free(curpath);
-	}
-
-	/*
-	 * if devt specified on the cmd line,
-	 * only consider affiliated paths
-	 */
-	if (conf->dev_type == DEV_DEVT && !devt2devname(buff, conf->dev)) {
-		dbg("limited scope = %s", conf->dev);
-		curpath = zalloc(sizeof (struct path));
-		if(safe_sprintf(curpath->dev, "%s", buff)) {
-			fprintf(stderr, "curpath->dev too small\n");
-			exit(1);
-		}
-		if (devinfo(curpath))
-			return 1;
-
-		memcpy(refwwid, curpath->wwid, WWID_SIZE);
-		free(curpath);
-	}
+		condlog(3, "limited scope = %s", conf->dev);
+		basename(conf->dev, buff);
+		pp = find_path_by_dev(pathvec, buff);
 		
-	if(safe_sprintf(path, "%s/block", sysfs_path)) {
-		fprintf(stderr, "path too small\n");
-		exit(1);
+		if (!pp) {
+			pp = alloc_path();
+
+			if (!pp)
+				return NULL;
+
+			strncpy(pp->dev, buff, FILE_NAME_SIZE);
+
+			if (pathinfo(pp, conf->hwtable, DI_SYSFS | DI_WWID))
+				return NULL;
+
+			if (store_path(pathvec, pp)) {
+				free_path(pp);
+				return NULL;
+			}
+		}
+
+		refwwid = MALLOC(WWID_SIZE);
+
+		if (!refwwid)
+			return NULL;
+
+		memcpy(refwwid, pp->wwid, WWID_SIZE);
+		return refwwid;
 	}
-	sdir = sysfs_open_directory(path);
-	sysfs_read_directory(sdir);
 
-	dlist_for_each_data(sdir->subdirs, devp, struct sysfs_directory) {
-		if (blacklist(conf->blist, devp->name))
-			continue;
+	if (conf->dev_type == DEV_DEVT) {
+		condlog(3, "limited scope = %s", conf->dev);
+		pp = find_path_by_devt(pathvec, conf->dev);
+		
+		if (!pp) {
+			pp = alloc_path();
 
-		sysfs_read_directory(devp);
+			if (!pp)
+				return NULL;
 
-		if (devp->links == NULL)
-			continue;
+			devt2devname(conf->dev, buff);
 
-		dlist_for_each_data(devp->links, linkp, struct sysfs_link) {
-			if (!strncmp(linkp->name, "device", 6))
-				break;
+			if(safe_sprintf(pp->dev, "%s", buff)) {
+				fprintf(stderr, "pp->dev too small\n");
+				exit(1);
+			}
+			if (pathinfo(pp, conf->hwtable, DI_SYSFS | DI_WWID))
+				return NULL;
+			
+			if (store_path(pathvec, pp)) {
+				free_path(pp);
+				return NULL;
+			}
 		}
-		if (linkp == NULL) {
-			continue;
-		}
-		basename(devp->path, buff);
-		curpath = zalloc(sizeof(struct path));
 
-		if(safe_sprintf(curpath->dev, "%s", buff)) {
-			fprintf(stderr, "curpath->dev too small\n");
-			exit(1);
-		}
-		if (devinfo(curpath)) {
-			fprintf(stderr, "pb getting path info, free %s\n",
-					curpath->dev);
-			free (curpath);
-			continue;
-		}
-		if (memcmp(empty_buff, refwwid, WWID_SIZE) != 0 && 
-		    memcmp(curpath->wwid, refwwid, WWID_SIZE) != 0) {
-			dbg("skip path %s : out of scope", curpath->dev);
-			free(curpath);
-			continue;
-		}
-		curpath->hwe = find_hwe(conf->hwtable, curpath->vendor_id,
-						      curpath->product_id);
-		vector_alloc_slot(pathvec);
-		vector_set_slot(pathvec, curpath);
+		refwwid = MALLOC(WWID_SIZE);
+
+		if (!refwwid)
+			return NULL;
+		
+		memcpy(refwwid, pp->wwid, WWID_SIZE);
+		return refwwid;
 	}
-	sysfs_close_directory(sdir);
-	return 0;
+	if (conf->dev_type == DEV_DEVMAP) {
+		condlog(3, "limited scope = %s", conf->dev);
+		/*
+		 * may be an alias
+		 */
+		refwwid = get_mpe_wwid(conf->dev);
+
+		if (refwwid)
+			return refwwid;
+		
+		/*
+		 * or directly a wwid
+		 */
+		refwwid = MALLOC(WWID_SIZE);
+
+		if (!refwwid)
+			return NULL;
+
+		strncpy(refwwid, conf->dev, WWID_SIZE);
+		return refwwid;
+	}
+	return NULL;
 }
 
-/*
- * print_path styles
- */
-#define PRINT_PATH_ALL		0
-#define PRINT_PATH_SHORT	1
-
 static void
-print_path (struct path * pp, int style)
+print_path (struct path * pp, char * style)
 {
-	if (style != PRINT_PATH_SHORT && pp->wwid)
-		printf ("%s ", pp->wwid);
-	else
-		printf ("  \\_ ");
+	char buff[MAX_LINE_LEN];
 
-	printf("%i:%i:%i:%i ",
-	       pp->sg_id.host_no,
-	       pp->sg_id.channel,
-	       pp->sg_id.scsi_id,
-	       pp->sg_id.lun);
-
-	if (pp->dev)
-		printf("%-4s ", pp->dev);
-
-	if (pp->dev_t)
-		printf("%-7s ", pp->dev_t);
-
-	switch (pp->state) {
-	case PATH_UP:
-		printf("[ready ]");
-		break;
-	case PATH_DOWN:
-		printf("[faulty]");
-		break;
-	case PATH_SHAKY:
-		printf("[shaky ]");
-		break;
-	default:
-		printf("[undef ]");
-		break;
-	}
-	switch (pp->dmstate) {
-	case PSTATE_ACTIVE:
-		printf("[active]");
-		break;
-	case PSTATE_FAILED:
-		printf("[failed]");
-		break;
-	default:
-		break;
-	}
-	if (pp->claimed)
-		printf("[claimed]");
-
-	if (style != PRINT_PATH_SHORT && pp->product_id)
-		printf("[%.16s]", pp->product_id);
-
-	fprintf(stdout, "\n");
+	snprint_path(&buff[0], MAX_LINE_LEN, style, pp, &pl);
+	printf("%s", buff);
 }
 
-#if DEBUG
 static void
 print_map (struct multipath * mpp)
 {
 	if (mpp->size && mpp->params)
-		printf("0 %lu %s %s\n",
+		printf("0 %llu %s %s\n",
 			 mpp->size, DEFAULT_TARGET, mpp->params);
 	return;
 }
@@ -239,21 +170,11 @@ static void
 print_all_paths (vector pathvec)
 {
 	int i;
-	char empty_buff[WWID_SIZE];
 	struct path * pp;
 
-	/* initialize a cmp 0-filled buffer */
-	memset(empty_buff, 0, WWID_SIZE);
-
-	vector_foreach_slot (pathvec, pp, i) {
-		/* leave out paths with incomplete devinfo */
-		if (memcmp(empty_buff, pp->wwid, WWID_SIZE) == 0)
-			continue;
-
-		print_path(pp, PRINT_PATH_ALL);
-	}
+	vector_foreach_slot (pathvec, pp, i)
+		print_path(pp, PRINT_PATH_LONG);
 }
-#endif
 
 static void
 print_mp (struct multipath * mpp)
@@ -296,14 +217,14 @@ print_mp (struct multipath * mpp)
 
 	printf("\n");
 
-	if (mpp->size < 2000)
-		printf("[size=%lu kB]", mpp->size / 2);
-	else if (mpp->size < (2000 * 1024))
-		printf("[size=%lu MB]", mpp->size / 2 / 1024);
-	else if (mpp->size < (2000 * 1024 * 1024))
-		printf("[size=%lu GB]", mpp->size / 2 / 1024 / 1024);
+	if (mpp->size < (1 << 11))
+		printf("[size=%llu kB]", mpp->size >> 1);
+	else if (mpp->size < (1 << 21))
+		printf("[size=%llu MB]", mpp->size >> 11);
+	else if (mpp->size < (1 << 31))
+		printf("[size=%llu GB]", mpp->size >> 21);
 	else
-		printf("[size=%lu TB]", mpp->size / 2 / 1024 / 1024 / 1024);
+		printf("[size=%llu TB]", mpp->size >> 31);
 
 	if (mpp->features)
 		printf("[features=\"%s\"]", mpp->features);
@@ -319,8 +240,17 @@ print_mp (struct multipath * mpp)
 	vector_foreach_slot (mpp->pg, pgp, j) {
 		printf("\\_ ");
 
-		if (mpp->selector)
+		if (mpp->selector) {
 			printf("%s ", mpp->selector);
+#if 0
+			/* align to path status info */
+			for (i = pl.hbtl_len + pl.dev_len + pl.dev_t_len + 4;
+			     i > strlen(mpp->selector); i--)
+				printf(" ");
+#endif
+		}
+		if (pgp->priority)
+			printf("[prio=%i]", pgp->priority);
 
 		switch (pgp->status) {
 		case PGSTATE_ENABLED:
@@ -335,25 +265,43 @@ print_mp (struct multipath * mpp)
 		default:
 			break;
 		}
-		if (mpp->nextpg && mpp->nextpg == j + 1)
-			printf("[first]");
-
 		printf("\n");
 
 		vector_foreach_slot (pgp->paths, pp, i)
-			print_path(pp, PRINT_PATH_SHORT);
+			print_path(pp, PRINT_PATH_INDENT);
 	}
 	printf("\n");
 }
 
+static int
+filter_pathvec (vector pathvec, char * refwwid)
+{
+	int i;
+	struct path * pp;
+
+	if (!refwwid || !strlen(refwwid))
+		return 0;
+
+	vector_foreach_slot (pathvec, pp, i) {
+		if (memcmp(pp->wwid, refwwid, WWID_SIZE) != 0) {
+			condlog(3, "skip path %s : out of scope", pp->dev);
+			free_path(pp);
+			vector_del_slot(pathvec, i);
+			i--;
+		}
+	}
+	return 0;
+}
+
 /*
  * Transforms the path group vector into a proper device map string
  */
-void
+int
 assemble_map (struct multipath * mp)
 {
 	int i, j;
 	int shift, freechar;
+	int minio;
 	char * p;
 	struct pathgroup * pgp;
 	struct path * pp;
@@ -367,7 +315,7 @@ assemble_map (struct multipath * mp)
 
 	if (shift >= freechar) {
 		fprintf(stderr, "mp->params too small\n");
-		exit(1);
+		return 1;
 	}
 	p += shift;
 	freechar -= shift;
@@ -378,17 +326,22 @@ assemble_map (struct multipath * mp)
 				 VECTOR_SIZE(pgp->paths));
 		if (shift >= freechar) {
 			fprintf(stderr, "mp->params too small\n");
-			exit(1);
+			return 1;
 		}
 		p += shift;
 		freechar -= shift;
 
 		vector_foreach_slot (pgp->paths, pp, j) {
+			minio = conf->minio;
+			
+			if (mp->rr_weight == RR_WEIGHT_PRIO && pp->priority)
+				minio *= pp->priority;
+
 			shift = snprintf(p, freechar, " %s %d",
-					 pp->dev_t, conf->minio);
+					 pp->dev_t, minio);
 			if (shift >= freechar) {
 				fprintf(stderr, "mp->params too small\n");
-				exit(1);
+				return 1;
 			}
 			p += shift;
 			freechar -= shift;
@@ -396,35 +349,40 @@ assemble_map (struct multipath * mp)
 	}
 	if (freechar < 1) {
 		fprintf(stderr, "mp->params too small\n");
-		exit(1);
+		return 1;
 	}
 	snprintf(p, 1, "\n");
-#if DEBUG
-	print_map(mp);
-#endif
+
+	if (conf->verbosity > 2)
+		print_map(mp);
+
+	return 0;
 }
 
 static int
 setup_map (struct multipath * mpp)
 {
 	struct path * pp;
-	struct pathgroup * pgp;
-	int i, j;
-	int highest = 0;
+	int i;
 
 	/*
 	 * don't bother if devmap size is unknown
 	 */
-	if (mpp->size <= 0)
+	if (mpp->size <= 0) {
+		condlog(3, "%s devmap size is unknown", mpp->alias);
 		return 1;
+	}
 
 	/*
 	 * don't bother if a constituant path is claimed
-	 * FIXME : claimed detection broken, always unclaimed for now
+	 * (not by the device mapper driver)
 	 */
-	vector_foreach_slot (mpp->paths, pp, i)
-		if (pp->claimed)
+	vector_foreach_slot (mpp->paths, pp, i) {
+		if (pp->claimed && pp->dmstate == PSTATE_UNDEF) {
+			condlog(3, "%s claimed", pp->dev);
 			return 1;
+		}
+	}
 
 	/*
 	 * properties selectors
@@ -433,6 +391,7 @@ setup_map (struct multipath * mpp)
 	select_selector(mpp);
 	select_features(mpp);
 	select_hwhandler(mpp);
+	select_rr_weight(mpp);
 
 	/*
 	 * apply selected grouping policy to valid paths
@@ -456,32 +415,26 @@ setup_map (struct multipath * mpp)
 	default:
 		break;
 	}
+
 	if (mpp->pg == NULL) {
-		dbg("pgpolicy failed to produce a ->pg vector");
+		condlog(3, "pgpolicy failed to produce a pg vector");
 		return 1;
 	}
 
 	/*
 	 * ponders each path group and determine highest prio pg
+	 * to switch over (default to first)
 	 */
-	mpp->nextpg = 1;
-	vector_foreach_slot (mpp->pg, pgp, i) {
-		vector_foreach_slot (pgp->paths, pp, j) {
-			pgp->id ^= (long)pp;
-			if (pp->state != PATH_DOWN)
-				pgp->priority += pp->priority;
-		}
-		if (pgp->priority > highest) {
-			highest = pgp->priority;
-			mpp->nextpg = i + 1;
-		}
-	}
+	select_path_group(mpp);
 
 	/*
 	 * transform the mp->pg vector of vectors of paths
 	 * into a mp->params strings to feed the device-mapper
 	 */
-	assemble_map(mpp);
+	if (assemble_map(mpp)) {
+		condlog(3, "problem assembing map");
+		return 1;
+	}
 	return 0;
 }
 
@@ -500,45 +453,16 @@ pathcount (struct multipath * mpp, int state)
 	return count;
 }
 
-/*
- * detect if a path is in the map we are about to create but not in the
- * current one (triggers a valid reload)
- * if a path is in the current map but not in the one we are about to create,
- * don't reload : it may come back latter so save the reload burden
- */
-static int
-pgcmp2 (struct multipath * mpp, struct multipath * cmpp)
+static void
+compute_pgid(struct pathgroup * pgp)
 {
-	int i, j, k, l;
-	struct pathgroup * pgp;
-	struct pathgroup * cpgp;
 	struct path * pp;
-	struct path * cpp;
-	int found = 0;
+	int i;
 
-	vector_foreach_slot (mpp->pg, pgp, i) {
-		vector_foreach_slot (pgp->paths, pp, j) {
-			vector_foreach_slot (cmpp->pg, cpgp, k) {
-				vector_foreach_slot (cpgp->paths, cpp, l) {
-					if (pp == cpp) {
-						found = 1;
-						break;
-					}
-				}
-				if (found)
-					break;
-			}
-			if (found) {
-				found = 0;
-				break;
-			} else
-				return 1;
-		}
-	}
-	return 0;
+	vector_foreach_slot (pgp->paths, pp, i)
+		pgp->id ^= (long)pp;
 }
 
-#if 0
 static int
 pgcmp (struct multipath * mpp, struct multipath * cmpp)
 {
@@ -548,6 +472,8 @@ pgcmp (struct multipath * mpp, struct multipath * cmpp)
 	int r = 0;
 
 	vector_foreach_slot (mpp->pg, pgp, i) {
+		compute_pgid(pgp);
+
 		vector_foreach_slot (cmpp->pg, cpgp, j) {
 			if (pgp->id == cpgp->id) {
 				r = 0;
@@ -560,7 +486,6 @@ pgcmp (struct multipath * mpp, struct multipath * cmpp)
 	}
 	return r;
 }
-#endif
 
 static void
 select_action (struct multipath * mpp, vector curmp)
@@ -570,49 +495,56 @@ select_action (struct multipath * mpp, vector curmp)
 	cmpp = find_mp(curmp, mpp->alias);
 
 	if (!cmpp) {
+		cmpp = find_mp_by_wwid(curmp, mpp->wwid);
+
+		if (cmpp && !conf->dry_run) {
+			condlog(2, "remove: %s (dup of %s)",
+				cmpp->alias, mpp->alias);
+			dm_flush_map(cmpp->alias, DEFAULT_TARGET);
+		}
 		mpp->action = ACT_CREATE;
 		return;
 	}
 	if (pathcount(mpp, PATH_UP) == 0) {
-		dbg("no good path");
+		condlog(3, "no good path");
 		mpp->action = ACT_NOTHING;
 		return;
 	}
 	if (cmpp->size != mpp->size) {
-		dbg("size different than current");
+		condlog(3, "size different than current");
 		mpp->action = ACT_RELOAD;
 		return;
 	}
 	if (strncmp(cmpp->features, mpp->features,
 		    strlen(mpp->features))) {
-		dbg("features different than current");
+		condlog(3, "features different than current");
 		mpp->action =  ACT_RELOAD;
 		return;
 	}
 	if (strncmp(cmpp->hwhandler, mpp->hwhandler,
 		    strlen(mpp->hwhandler))) {
-		dbg("hwhandler different than current");
+		condlog(3, "hwhandler different than current");
 		mpp->action = ACT_RELOAD;
 		return;
 	}
 	if (strncmp(cmpp->selector, mpp->selector,
 		    strlen(mpp->selector))) {
-		dbg("selector different than current");
+		condlog(3, "selector different than current");
 		mpp->action = ACT_RELOAD;
 		return;
 	}
 	if (VECTOR_SIZE(cmpp->pg) != VECTOR_SIZE(mpp->pg)) {
-		dbg("different number of PG");
+		condlog(3, "different number of PG");
 		mpp->action = ACT_RELOAD;
 		return;
 	}
-	if (pgcmp2(mpp, cmpp)) {
-		dbg("different path group topology");
+	if (pgcmp(mpp, cmpp)) {
+		condlog(3, "different path group topology");
 		mpp->action = ACT_RELOAD;
 		return;
 	}
 	if (cmpp->nextpg != mpp->nextpg) {
-		dbg("nextpg different than current");
+		condlog(3, "nextpg different than current");
 		mpp->action = ACT_SWITCHPG;
 		return;
 	}
@@ -620,39 +552,6 @@ select_action (struct multipath * mpp, vector curmp)
 	return;
 }
 
-static int
-dm_switchgroup(char * mapname, int index)
-{
-	int r = 0;
-	struct dm_task *dmt;
-	char str[24];
-
-	if (!(dmt = dm_task_create(DM_DEVICE_TARGET_MSG)))
-		return 0;
-
-	if (!dm_task_set_name(dmt, mapname))
-		goto out;
-
-	if (!dm_task_set_sector(dmt, 0))
-		goto out;
-
-	snprintf(str, 24, "switch_group %i\n", index);
-	dbg("message %s 0 %s", mapname, str);
-
-	if (!dm_task_set_message(dmt, str))
-		goto out;
-
-	if (!dm_task_run(dmt))
-		goto out;
-
-	r = 1;
-
-	out:
-	dm_task_destroy(dmt);
-
-	return r;
-}
-
 static int
 reinstate_paths (struct multipath * mpp)
 {
@@ -668,7 +567,9 @@ reinstate_paths (struct multipath * mpp)
 				continue;
 
 			if (pp->dmstate == PSTATE_FAILED) {
-				dm_reinstate(mpp->alias, pp->dev_t);
+				if (dm_reinstate(mpp->alias, pp->dev_t))
+					condlog(0, "error reinstating %s",
+						pp->dev);
 			}
 		}
 	}
@@ -678,7 +579,7 @@ reinstate_paths (struct multipath * mpp)
 static int
 domap (struct multipath * mpp)
 {
-	int op;
+	int op = ACT_NOTHING;
 	int r = 0;
 
 	print_mp(mpp);
@@ -686,10 +587,14 @@ domap (struct multipath * mpp)
 	/*
 	 * last chance to quit before touching the devmaps
 	 */
-	if (conf->dry_run || mpp->action == ACT_NOTHING)
+	if (conf->dry_run)
 		return 0;
 
-	if (mpp->action == ACT_SWITCHPG) {
+	switch (mpp->action) {
+	case ACT_NOTHING:
+		return 0;
+
+	case ACT_SWITCHPG:
 		dm_switchgroup(mpp->alias, mpp->nextpg);
 		/*
 		 * we may have avoided reinstating paths because there where in
@@ -698,12 +603,18 @@ domap (struct multipath * mpp)
 		 */
 		reinstate_paths(mpp);
 		return 0;
-	}
-	if (mpp->action == ACT_CREATE)
-		op = DM_DEVICE_CREATE;
 
-	if (mpp->action == ACT_RELOAD)
+	case ACT_CREATE:
+		op = DM_DEVICE_CREATE;
+		break;
+
+	case ACT_RELOAD:
 		op = DM_DEVICE_RELOAD;
+		break;
+
+	default:
+		break;
+	}
 
 		
 	/*
@@ -731,7 +642,7 @@ domap (struct multipath * mpp)
 	return r;
 }
 
-static void
+static int
 coalesce_paths (vector curmp, vector pathvec)
 {
 	int k, i;
@@ -745,8 +656,9 @@ coalesce_paths (vector curmp, vector pathvec)
 	vector_foreach_slot (pathvec, pp1, k) {
 		/* skip this path for some reason */
 
-		/* 1. if path has no unique id */
-		if (memcmp(empty_buff, pp1->wwid, WWID_SIZE) == 0)
+		/* 1. if path has no unique id or wwid blacklisted */
+		if (memcmp(empty_buff, pp1->wwid, WWID_SIZE) == 0 ||
+		    blacklist(conf->blist, pp1->wwid))
 			continue;
 
 		/* 2. if path already coalesced */
@@ -756,26 +668,28 @@ coalesce_paths (vector curmp, vector pathvec)
 		/*
 		 * at this point, we know we really got a new mp
 		 */
-		mpp = zalloc(sizeof(struct multipath));
+		mpp = alloc_multipath();
+
+		if (!mpp)
+			return 1;
 
 		mpp->mpe = find_mpe(pp1->wwid);
 		mpp->hwe = pp1->hwe;
 		select_alias(mpp);
 
-		if (conf->dev_type == DEV_DEVMAP &&
-		    (mpp->alias &&
-		    0 != strncmp(mpp->alias, conf->dev, FILE_NAME_SIZE)) &&
-		    0 != strncmp(pp1->wwid, conf->dev, FILE_NAME_SIZE)) {
-			free(mpp);
-			continue;
-		}
 		pp1->mpp = mpp;
 		strcpy(mpp->wwid, pp1->wwid);
 		mpp->size = pp1->size;
-
 		mpp->paths = vector_alloc();
-		vector_alloc_slot (mpp->paths);
-		vector_set_slot (mpp->paths, pp1);
+
+		if (pp1->priority < 0)
+			mpp->action = ACT_NOTHING;
+
+		if (!mpp->paths)
+			return 1;
+		
+		if (store_path(mpp->paths, pp1))
+			return 1;
 
 		for (i = k + 1; i < VECTOR_SIZE(pathvec); i++) {
 			pp2 = VECTOR_SLOT(pathvec, i);
@@ -789,110 +703,38 @@ coalesce_paths (vector curmp, vector pathvec)
 				/*
 				 * ouch, avoid feeding that to the DM
 				 */
-				dbg("path size mismatch : discard %s",
+				condlog(3, "path size mismatch : discard %s",
 				     mpp->wwid);
 				mpp->action = ACT_NOTHING;
 			}
-			vector_alloc_slot(mpp->paths);
-			vector_set_slot(mpp->paths, VECTOR_SLOT(pathvec, i));
-		}
-		if (mpp) {
-			setup_map(mpp);
+			if (pp2->priority < 0)
+				mpp->action = ACT_NOTHING;
 
-			if (mpp->action == ACT_UNDEF)
-				select_action(mpp, curmp);
-
-			domap(mpp);
-			free_multipath(mpp);
+			if (store_path(mpp->paths, pp2))
+				return 1;
 		}
+		if (setup_map(mpp)) {
+			free_multipath(mpp, KEEP_PATHS);
+			continue;
+		}
+		condlog(3, "action preset to %i", mpp->action);
+
+		if (mpp->action == ACT_UNDEF)
+			select_action(mpp, curmp);
+
+		condlog(3, "action set to %i", mpp->action);
+
+		domap(mpp);
+		free_multipath(mpp, KEEP_PATHS);
 	}
-}
-
-static int
-dm_get_maps (vector mp, char * type)
-{
-	struct multipath * mpp;
-	int r = 0;
-	struct dm_task *dmt;
-	struct dm_names *names;
-	unsigned next = 0;
-	unsigned long length;
-	char *params;
-	char *status;
-
-	if (!(dmt = dm_task_create (DM_DEVICE_LIST)))
-		return 0;
-
-	if (!dm_task_run (dmt))
-		goto out;
-
-	if (!(names = dm_task_get_names (dmt)))
-		goto out;
-
-	if (!names->dev)
-		goto out;
-
-	do {
-		if (dm_type(names->name, DEFAULT_TARGET)) {
-			dm_get_map(names->name, &length, &params);
-			dm_get_status(names->name, &status);
-			mpp = zalloc(sizeof(struct multipath));
-
-			if (!mpp) {
-				r = 1;
-				goto out;
-			}
-			mpp->size = length;
-			mpp->alias = zalloc(strlen(names->name) + 1);
-			strncat(mpp->alias, names->name, strlen(names->name));
-			strncat(mpp->params, params, PARAMS_SIZE);
-			strncat(mpp->status, status, PARAMS_SIZE);
-
-			vector_alloc_slot(mp);
-			vector_set_slot(mp, mpp);
-			mpp = NULL;
-		}
-                next = names->next;
-                names = (void *) names + next;
-	} while (next);
-
-	out:
-	dm_task_destroy (dmt);
-	return r;
-}
-
-static void
-signal_daemon (void)
-{
-	FILE *file;
-	pid_t pid;
-	char *buf;
-
-	buf = malloc (8);
-
-	file = fopen (DEFAULT_PIDFILE, "r");
-
-	if (!file) {
-		if (conf->verbosity > 0)
-			fprintf (stderr, "cannot signal daemon, "
-					 "pidfile not found\n");
-		return;
-	}
-
-	buf = fgets (buf, 8, file);
-	fclose (file);
-
-	pid = (pid_t) atol (buf);
-	free (buf);
-
-	kill (pid, SIGHUP);
+	return 0;
 }
 
 static void
 usage (char * progname)
 {
 	fprintf (stderr, VERSION_STRING);
-	fprintf (stderr, "Usage: %s\t[-v level] [-d] [-S]\n",
+	fprintf (stderr, "Usage: %s\t[-v level] [-d] [-l|-ll|-f|-F]\n",
 		progname);
 	fprintf (stderr,
 		"\t\t\t[-p failover|multibus|group_by_serial|group_by_prio]\n" \
@@ -901,9 +743,12 @@ usage (char * progname)
 		"\t-v level\tverbosty level\n" \
 		"\t   0\t\t\tno output\n" \
 		"\t   1\t\t\tprint created devmap names only\n" \
-		"\t   2\t\t\tprint all paths and multipaths\n" \
+		"\t   2\t\t\tdefault verbosity\n" \
+		"\t   3\t\t\tprint debug information\n" \
 		"\t-d\t\tdry run, do not create or update devmaps\n" \
-		"\t-S\t\tinhibit signal sending to multipathd\n"
+		"\t-l\t\tshow multipath topology (sysfs and DM info)\n" \
+		"\t-ll\t\tshow multipath topology (maximum info)\n" \
+		"\t-F\t\tflush a multipath device map\n" \
 		"\t-F\t\tflush all multipath device maps\n" \
 		"\t-p policy\tforce all maps to specified policy :\n" \
 		"\t   failover\t\t1 path per priority group\n" \
@@ -921,7 +766,7 @@ usage (char * progname)
 }
 
 static int
-get_current_paths (struct multipath * mpp, vector pathvec)
+update_paths (struct multipath * mpp)
 {
 	int i, j;
 	struct pathgroup * pgp;
@@ -929,36 +774,60 @@ get_current_paths (struct multipath * mpp, vector pathvec)
 
 	vector_foreach_slot (mpp->pg, pgp, i) {
 		vector_foreach_slot (pgp->paths, pp, j) {
-			if (strlen(pp->dev) == 0) {
+			if (!strlen(pp->dev)) {
 				devt2devname(pp->dev, pp->dev_t);
-				devinfo(pp);
+				pathinfo(pp, conf->hwtable,
+					 DI_SYSFS | DI_CHECKER | \
+					 DI_SERIAL | DI_PRIO);
+				continue;
 			}
 			if (pp->state == PATH_UNCHECKED)
-				pp->state = pp->checkfn(pp->dev_t, NULL, NULL);
+				pathinfo(pp, conf->hwtable, DI_CHECKER);
+
+			if (!pp->priority)
+				pathinfo(pp, conf->hwtable, DI_PRIO);
 		}
 	}
 	return 0;
 }
 
 static int
-get_current_mp (vector curmp, vector pathvec)
+get_dm_mpvec (vector curmp, vector pathvec, char * refwwid)
 {
 	int i;
 	struct multipath * mpp;
+	char * wwid;
 
-	dm_get_maps(curmp, DEFAULT_TARGET);
+	if (dm_get_maps(curmp, DEFAULT_TARGET))
+		return 1;
 
 	vector_foreach_slot (curmp, mpp, i) {
-		if (conf->dev_type == DEV_DEVMAP && (
-		    (mpp->alias &&
-		    0 != strncmp(mpp->alias, conf->dev, FILE_NAME_SIZE)) &&
-		    0 != strncmp(mpp->wwid, conf->dev, FILE_NAME_SIZE)))
-			continue;
+		wwid = get_mpe_wwid(mpp->alias);
 
-		dbg("params = %s", mpp->params);
-		dbg("status = %s", mpp->status);
+		if (wwid) {
+			/* out of specified scope */
+			if (refwwid && strncmp(wwid, refwwid, WWID_SIZE))
+				continue;
+			wwid = NULL;
+		}
+
+		condlog(3, "params = %s", mpp->params);
+		condlog(3, "status = %s", mpp->status);
+
+		/* will set mpp->wwid */
 		disassemble_map(pathvec, mpp->params, mpp);
-		get_current_paths(mpp, pathvec);
+
+		/*
+		 * disassemble_map may have added new paths to pathvec.
+		 * If not in "fast list mode", we need to fetch information
+		 * about them
+		 */
+		if (conf->list != 1)
+			update_paths(mpp);
+
+		if (conf->list > 1)
+			select_path_group(mpp);
+
 		disassemble_status(mpp->status, mpp);
 
 		if (conf->list)
@@ -973,44 +842,31 @@ get_current_mp (vector curmp, vector pathvec)
 int
 main (int argc, char *argv[])
 {
-	vector curmp;
-	vector pathvec;
+	vector curmp = NULL;
+	vector pathvec = NULL;
 	int i;
+	int di_flag = 0;
 	int arg;
 	extern char *optarg;
 	extern int optind;
+	char * refwwid = NULL;
 
-	if (dm_prereq(DEFAULT_TARGET, 1, 0, 3)) {
-		fprintf(stderr, "device mapper prerequisites not met.\n");
+	if (getuid() != 0) {
+		fprintf(stderr, "need to be root\n");
 		exit(1);
 	}
 
-	/*
-	 * alloc config struct
-	 */
-	conf = zalloc(sizeof(struct config));
-				
-	/*
-	 * internal defaults
-	 */
-	conf->list = 0;
-	conf->dry_run = 0;		/* 1 == Do not Create/Update devmaps */
-	conf->verbosity = 1;
-	conf->pgpolicy_flag = 0;	/* do not override defaults */
-	conf->signal = 1;		/* 1 == Send a signal to multipathd */
-	conf->dev = NULL;
-	conf->dev_type = DEV_NONE;
-	conf->default_selector = NULL;
-	conf->default_selector_args = 0;
-	conf->default_pgpolicy = 0;
-	conf->mptable = NULL;
-	conf->hwtable = NULL;
-	conf->blist = NULL;
-	conf->default_features = NULL;
-	conf->default_hwhandler = NULL;
-	conf->minio = 1000;
+	if (dm_prereq(DEFAULT_TARGET, 1, 0, 3))
+		exit(1);
 
-	while ((arg = getopt(argc, argv, ":qdlFSi:v:p:")) != EOF ) {
+	if (sysfs_get_mnt_path(sysfs_path, FILE_NAME_SIZE)) {
+		condlog(0, "multipath tools need sysfs mounted");
+		exit(1);
+	}
+	if (load_config(DEFAULT_CONFIGFILE))
+		exit(1);
+
+	while ((arg = getopt(argc, argv, ":qdl::Ffi:M:v:p:")) != EOF ) {
 		switch(arg) {
 		case 1: printf("optarg : %s\n",optarg);
 			break;
@@ -1023,19 +879,25 @@ main (int argc, char *argv[])
 			break;
 		case 'd':
 			conf->dry_run = 1;
-			conf->signal = 0;
+			break;
+		case 'f':
+			conf->remove = 1;
 			break;
 		case 'F':
-			dm_flush_maps(DEFAULT_TARGET);
-			goto out;
+			conf->remove = 2;
 			break;
 		case 'l':
 			conf->list = 1;
 			conf->dry_run = 1;
-			conf->signal = 0;
+
+			if (optarg && !strncmp(optarg, "l", 1))
+				conf->list++;
+
 			break;
-		case 'S':
-			conf->signal = 0;
+		case 'M':
+#if _DEBUG_
+			debug = atoi(optarg);
+#endif
 			break;
 		case 'p':
 			conf->pgpolicy_flag = get_pgpolicy_id(optarg);
@@ -1054,8 +916,12 @@ main (int argc, char *argv[])
 			usage(argv[0]);
 		}
 	}        
-	if (optind<argc) {
-		conf->dev = zalloc(FILE_NAME_SIZE);
+	if (optind < argc) {
+		conf->dev = MALLOC(FILE_NAME_SIZE);
+
+		if (!conf->dev)
+			goto out;
+
 		strncpy(conf->dev, argv[optind], FILE_NAME_SIZE);
 
 		if (filepresent(conf->dev))
@@ -1067,6 +933,19 @@ main (int argc, char *argv[])
 
 	}
 
+	if (conf->remove == FLUSH_ONE) {
+		if (conf->dev_type == DEV_DEVMAP)
+			dm_flush_map(conf->dev, DEFAULT_TARGET);
+		else
+			condlog(0, "must provide a map name to remove");
+
+		goto out;
+	}
+	else if (conf->remove == FLUSH_ALL) {
+		dm_flush_maps(DEFAULT_TARGET);
+		goto out;
+	}
+
 	/*
 	 * allocate core vectors to store paths and multipaths
 	 */
@@ -1074,82 +953,75 @@ main (int argc, char *argv[])
 	pathvec = vector_alloc();
 
 	if (!curmp || !pathvec) {
-		fprintf(stderr, "can not allocate memory\n");
-		exit(1);
+		condlog(0, "can not allocate memory");
+		goto out;
 	}
 
-	/*
-	 * read the config file
-	 */
-	if (filepresent(DEFAULT_CONFIGFILE))
-		init_data (DEFAULT_CONFIGFILE, init_keywords);
-	
-	/*
-	 * fill the voids left in the config file
-	 */
-	if (conf->hwtable == NULL) {
-		conf->hwtable = vector_alloc();
-		setup_default_hwtable(conf->hwtable);
-	}
-	if (conf->blist == NULL) {
-		conf->blist = vector_alloc();
-		setup_default_blist(conf->blist);
-	}
-	if (conf->mptable == NULL)
-		conf->mptable = vector_alloc();
-
-	if (conf->default_selector == NULL)
-		conf->default_selector = DEFAULT_SELECTOR;
-
-	if (conf->udev_dir == NULL)
-		conf->udev_dir = DEFAULT_UDEVDIR;
-
-	if (conf->default_getuid == NULL)
-		conf->default_getuid = DEFAULT_GETUID;
-
-	if (conf->default_features == NULL)
-		conf->default_features = DEFAULT_FEATURES;
-
-	if (conf->default_hwhandler == NULL)
-		conf->default_hwhandler = DEFAULT_HWHANDLER;
-
 	/*
 	 * if we have a blacklisted device parameter, exit early
 	 */
 	if (conf->dev && blacklist(conf->blist, conf->dev))
-		exit(0);
+		goto out;
 	
+	condlog(3, "load path identifiers cache");
+	cache_load(pathvec);
+
+	if (conf->verbosity > 2) {
+		fprintf(stdout, "#\n# all paths in cache :\n#\n");
+		print_all_paths(pathvec);
+	}
+
 	/*
 	 * get a path list
 	 */
-	if (get_pathvec_sysfs(pathvec) || VECTOR_SIZE(pathvec) == 0)
-		exit(1);
+	if (conf->dev)
+		di_flag = DI_WWID;
 
-	if (VECTOR_SIZE(pathvec) == 0 && conf->verbosity > 0) {
-		fprintf(stdout, "no path found\n");
-		exit(0);
+	if (conf->list > 1)
+		/* extended path info '-ll' */
+		di_flag |= DI_SYSFS | DI_CHECKER;
+	else if (conf->list)
+		/* minimum path info '-l' */
+		di_flag |= DI_SYSFS;
+	else
+		/* maximum info */
+		di_flag = DI_ALL;
+
+	if (path_discovery(pathvec, conf, di_flag) || VECTOR_SIZE(pathvec) == 0)
+		goto out;
+
+	if (conf->verbosity > 2) {
+		fprintf(stdout, "#\n# all paths :\n#\n");
+		print_all_paths(pathvec);
 	}
-#if DEBUG
-	fprintf(stdout, "#\n# all paths :\n#\n");
-	print_all_paths(pathvec);
-#endif
 
-	get_current_mp(curmp, pathvec);
+	refwwid = get_refwwid(pathvec);
+	get_path_layout(&pl, pathvec);
+
+	if (get_dm_mpvec(curmp, pathvec, refwwid))
+		goto out;
+
+	filter_pathvec(pathvec, refwwid);
 
 	if (conf->list)
 		goto out;
 
 	/*
-	 * group the paths as multipaths
+	 * core logic entry point
 	 */
 	coalesce_paths(curmp, pathvec);
 
 out:
-	/*
-	 * signal multipathd that new devmaps may have come up
-	 */
-	if (conf->signal)
-		signal_daemon();
-	
+	if (refwwid)
+		FREE(refwwid);
+
+	free_multipathvec(curmp, KEEP_PATHS);
+	free_pathvec(pathvec, FREE_PATHS);
+	free_config(conf);
+	dm_lib_release();
+	dm_lib_exit();
+#ifdef _DEBUG_
+	dbg_free_final(NULL);
+#endif
 	exit(0);
 }
