@@ -4,15 +4,15 @@
 #include <pthread.h>
 #include <sys/mman.h>
 
+#include <memory.h>
+
 #include "log_pthread.h"
 #include "log.h"
 
-void log_safe (int prio, char * fmt, ...)
+void log_safe (int prio, char * fmt, va_list ap)
 {
-	va_list ap;
-
 	pthread_mutex_lock(logq_lock);
-	va_start(ap, fmt);
+	//va_start(ap, fmt);
 	log_enqueue(prio, fmt, ap);
 	va_end(ap);
 	pthread_mutex_unlock(logq_lock);
